// Copyright Robin Schirrmeister and Simon Skilevic
// University Freiburg 2011

#include "./MulticriteriaExpandedDijkstra.h"
#include <algorithm>
#include <numeric>
#include <vector>
#include "./NodeDependent.h"
#include "./TransitPath.h"

MulticriteriaExpandedDijkstra::MulticriteriaExpandedDijkstra(TransitNetworkExpanded* transitNetwork)
{
  this->_transitGraph = transitNetwork;
  _timesFromStart.resize(_transitGraph->getNumNodes(), -1);
}


MulticriteriaExpandedDijkstra::~MulticriteriaExpandedDijkstra()
{
}

double MulticriteriaExpandedDijkstra::computeShortestPath(int startStopId, Time startTime,
                                           int endStopId, bool reconstructPath)
{
  optimalPaths.clear();
  transitShortestPath.clear();
  // reset vectors for algorithm
  labelSets.clear();
  labelSets.resize(_transitGraph->getNumNodes(), LabelSet(&labels, 0));
  // get the start node ids
  vector<int> startNodeIds;
  _transitGraph->getStartNodeIds(startStopId, startTime, &startNodeIds);

  // Cannot do Dijkstra without start ids..
  if (startNodeIds.size() == 0)
  {
    printf("No start nodes at id %d\n", startStopId);
    return -1;
  }
  // empty priority queue, then enter the starting nodes into the priority queue
  _openNodeQueue = std::priority_queue <pair<int, int>,
            vector<pair<int, int> >,
            std::greater<pair<int, int > > >();
  int currentTimeFromStart;

  vector <int> currentCriterias;
  currentCriterias.push_back(0);
  currentCriterias.push_back(0);

  Time nodeTime;
  for (size_t i = 0; i < startNodeIds.size(); i++)
  {
    nodeTime = _transitGraph->getNode(startNodeIds[i])->time;
    currentTimeFromStart = nodeTime.minusInSecondsDependent(startTime);
    currentCriterias[0] = currentTimeFromStart;
    currentCriterias[1] = 0;  // 0 Transfers
    bool flag = labelSets[startNodeIds[i]].AddLabel(
            Connection(_transitGraph->getTripId(startNodeIds[i])),
            currentCriterias, startNodeIds[i], -1, UNKNOWN, true, false, true, true);
    assert(flag == true);
    int sumOfCriterias = std::accumulate(currentCriterias.begin(),
                                            currentCriterias.end(),
                                            0.0);
    _openNodeQueue.push(std::pair<int, int> (sumOfCriterias,
                         labels.size() - 1));
  }
  int sumOfNewCriterias;
  int currentId;
  int currentLabelId;
  int currentTripId;
  vector<Edge>* edges;
  LabelSet endParitySet(&labels, 0);

  int newId;
  int newTripId;
  bool dijkstraFinished = false;
  size_t gepopt = 0;
  size_t exploredNodes = 0;
  while (!dijkstraFinished)
  {
    // get minimum of node queue
    currentLabelId = _openNodeQueue.top().second;
    currentId = labels[currentLabelId].NodeId;
    currentTripId = labels[currentLabelId].connection.lineId;
    NodeExpanded* currentNode = _transitGraph->getNode(currentId);
    _openNodeQueue.pop();
//    std::cout << currentId << " " <<  _openNodeQueue.size() << std::endl;
    gepopt++;

    // check if already explored
    if (labels[currentLabelId].removed == true)
    {
       dijkstraFinished = _openNodeQueue.empty();
       continue;
    }

    bool dominatedAtEndStation = false;
    if (endParitySet._expandedFrom.size() != 0)
    {
      if (!endParitySet.labelMustBeAdded(
              labels[currentLabelId].criterias[0],
              labels[currentLabelId].criterias[1],
              labels[currentLabelId].connection.lineId, 0, true, true, true))
      {
        labels[currentLabelId].removed = true;
        dominatedAtEndStation = true;
        dijkstraFinished = _openNodeQueue.empty();
      }
    }
    if (dominatedAtEndStation) continue;
    // get outgoing edges from this node
    edges = _transitGraph->getOutEdges(currentId);
    // compute all new distances in loop
    for (size_t i = 0; i < edges->size(); i++)
    {
      // get all new distances
      Edge& newEdge = (*edges)[i];
      newId = newEdge.otherNodeId;
      newTripId = _transitGraph->getTripId(newId);
      NodeExpanded* newNode = _transitGraph->getNode(newId);
      vector <int> newCriterias = labels[currentLabelId].criterias;

      // Time criteria
      newCriterias[0] += newEdge.time;
      // Transfer criteria
      if ((currentNode->type == TRANSFERNODE
       && newNode->type == DEPARTURENODE)
       && currentNode->stopId != startStopId)
        newCriterias[1] += 1;
      if (labelSets[newId].AddLabel(Connection(newTripId), newCriterias,
                               newId, currentLabelId, UNKNOWN, true, false, true, true))
      {
        sumOfNewCriterias = std::accumulate(newCriterias.begin(),
                                            newCriterias.end(),
                                            0.0);
        _openNodeQueue.push(std::pair<int, int> (sumOfNewCriterias,
                                                    labels.size() - 1));
      }
    }
    exploredNodes++;
    if (currentNode->stopId == endStopId)
    {
      bool fl = endParitySet.AddLabel(labels[currentLabelId].connection,
              labels[currentLabelId].criterias, currentId, currentLabelId,
              UNKNOWN, true, true, false, true);
      assert(fl);
    }
    dijkstraFinished = _openNodeQueue.empty();
  }
  if ( endParitySet._expandedFrom.size() == 0)
  {
    printf("shortest Path not found from %s(%d) to %s(%d)\n",
                                _transitGraph->getStopName(startStopId).c_str(),
                                startStopId,
                                _transitGraph->getStopName(endStopId).c_str(),
                                endStopId);
    return -1;
  }
  if (reconstructPath)
  {
    unordered_set<int> startSet;
    for (size_t i = 0; i < startNodeIds.size(); i++)
      startSet.insert(startNodeIds[i]);
    reconstructShortestPath(startTime, startSet, endParitySet);
  }
  return exploredNodes;
}

void MulticriteriaExpandedDijkstra::reconstructShortestPath(
                                  const Time& startTime,
                                  const unordered_set<int>& startIds,
                                  const LabelSet& endParitySet)
{
  NodeExpanded* node;
  int nodeId;
  int labelId;
  // reconstruct the path, leave out transfer nodes!
  for (size_t i = 0; i < endParitySet._expandedFrom.size(); i++)
    for (size_t j = 0; j < endParitySet._expandedFrom[i].size(); j++)
      if (!labels[endParitySet._expandedFrom[i][j]].removed)

      {
        // first add endnode, then go through other ndoes
        TransitPath newPath(startTime);
        labelId = endParitySet._expandedFrom[i][j];
        nodeId = labels[labelId].NodeId;
        node = _transitGraph->getNode(nodeId);
        double latitude = node->latitude;
        double longitude = node->longitude;
        // last node should be arrival node
        assert(node->type == ARRIVALNODE);
        Time arrivalTime, departureTime;
        arrivalTime = node->time;
        // add endnode to the path without departure time!
        newPath.addStation(TransitPath::TRANSITNODE, node->stopId,
                            latitude, longitude,
                            arrivalTime, departureTime,
                           _transitGraph->getTripId(nodeId));
        bool departureFound = false;
        bool arrivalFound = true;  // just for assert checks
        while (startIds.count(nodeId) == 0)
        {
          labelId = labels[labelId].expandedFromLabelId;
          nodeId = labels[labelId].NodeId;
          node = _transitGraph->getNode(nodeId);
          if (node->type == DEPARTURENODE)
          {
            assert(arrivalFound);
            departureTime = node->time;
            departureFound = true;
          }
          else if (node->type == ARRIVALNODE)
          {
            assert(departureFound);
            latitude = node->latitude;
            longitude = node->longitude;
            arrivalTime = node->time;
            newPath.addStation(TransitPath::TRANSITNODE,
                               node->stopId, latitude, longitude,
                               arrivalTime, departureTime,
                              _transitGraph->getTripId(nodeId));
            departureFound = false;
            arrivalFound = true;
          }
        }
        // we should have a departurenode at the start
        assert(departureFound);
        latitude = node->latitude;
        longitude = node->longitude;
        // add with start time as arrival time and dummy trip id -1
        newPath.addStation(TransitPath::TRANSITNODE,
                           node->stopId, latitude, longitude,
                           arrivalTime, departureTime,
                            _transitGraph->getTripId(nodeId));
        // reverse the path so that its in order from start to end,
        // and add it to the optimal paths
        newPath.reverse();
        optimalPaths.push_back(newPath);
      }
}

double MulticriteriaExpandedDijkstra::computeShortestPath(string startStop, Time time,
                                      string endStop, bool buildPath)

{
  int startStopId = _transitGraph->getStopId(startStop);
  int endStopId = _transitGraph->getStopId(endStop);
  if (startStopId == -1)
    std::cout << "Start Station " << startStop << " not found" << std::endl;
  if (endStopId == -1)
    std::cout << "End Station " << endStop << " not found" << std::endl;
  if (startStopId == -1 || endStopId == -1)
    return -1;
  return computeShortestPath(startStopId, time, endStopId, buildPath);
}
