// Copyright November 18, 2011 Robin Schirrmeister and Simon Skilevic
// Chair of Algorithms and Datastructures Uni Freiburg

#include "./RoadNetwork.h"
#include <map>
#include <string>
#include <sstream>
#include <iostream>
#include <vector>
#include <queue>
#include <math.h>

using std::cout;
using std::endl;

RoadNetwork::RoadNetwork() : Graph()
{
}


// _____________________________________________________________________________
void RoadNetwork::makeEdge(const string &line, FILE* osmFile,
                           map<string, int>* idMap,
                           const map<string, size_t>& speedTable)
{
  const int LINE_SIZE = 1000;
  char buf[LINE_SIZE];
  vector <size_t> wayNodes;
  size_t waySpeed = 0;
  size_t oneWay = false;
  string endWay = "</way>";
  size_t found = line.find(endWay);
  while ((found == string::npos) && (fgets(buf, LINE_SIZE, osmFile)))
  {
    string line = buf;
    size_t pos1 = line.find("<nd ref=");
    // Save way Node in a Vector if node discription readed
    if (pos1 != string::npos)
    {
      pos1+=9;
      size_t pos2 = line.find("\"", pos1);
      string idStr = line.substr(pos1, pos2 - pos1);
      wayNodes.push_back((*idMap)[idStr]);
    }
    else
    {
      // Extract type of road
      pos1 = line.find("<tag k=\"highway\" v=");
      if (pos1 != string::npos)
      {
        pos1+=20;
        size_t pos2 = line.find("\"", pos1);
        string roadType = line.substr(pos1, pos2 - pos1);
        if (speedTable.count(roadType) > 0)
          waySpeed = speedTable.at(roadType);
        else
        {
          return;  // TODO(Robin): check if using default speed would be better
                   // instead of just removing this edge altogether
        }
      }

      // Extract direction of road
      pos1 = line.find("<tag k=\"oneway\" v=");
      if (pos1 != string::npos)
      {
        pos1 += 19;
        size_t pos2 = line.find("\"", pos1);
        string roadDirection = line.substr(pos1, pos2 - pos1);
        if (roadDirection.compare("yes") == 0)
          oneWay = true;
      }
    }
    found = line.find(endWay);
  }

  if (waySpeed == 0) return;
  int time;
  int startId, endId;
  for (size_t i = 0; i < wayNodes.size() - 1; i++)
  {
    startId = wayNodes[i];
    endId = wayNodes[i + 1];
    time = getNode(startId)->straightLineDistanceTo(*(getNode(endId))) 
           / 5.0 * 60.0 * 60.0;
    // convert time to seconds and to walkingspeed
//    time *= 60 * 60;
//    time /= 5.0;
    if (time < 1)
      time = 1;
    addEdge(startId, endId, time);
    if (!oneWay)
    {
      addEdge(endId, startId, time);
    }
  }
}

void RoadNetwork::makeNode(const string &line, map<string, int>* idMap)
{
  // Extract Id
  size_t pos1 = line.find("id=");
  pos1+=4;
  size_t pos2 = line.find("\"", pos1);
  string idStr = line.substr(pos1, pos2 - pos1);
  (*idMap)[idStr] = _nodes->size();

  // Extract latitude
  pos1 = line.find("lat=");
  pos1+=5;
  pos2 = line.find("\"", pos1);
  string latStr = line.substr(pos1, pos2 - pos1);
  double lat = atof(latStr.c_str());

  // Extract longitude
  pos1 = line.find("lon=");
  pos1+=5;
  pos2 = line.find("\"", pos1);
  string lonStr = line.substr(pos1, pos2 - pos1);
  double lon = atof(lonStr.c_str());

  // Create new Node with extracted Data and add it to the Vector of Nodes
  NodeExpanded node((*idMap)[idStr], lat, lon);
  addNode(node);
}


// _____________________________________________________________________________
void RoadNetwork::readFromOsmFile(string fileName)
{
  const int LINE_SIZE = 1000;
  char buf[LINE_SIZE];
  FILE* osmFile = fopen(fileName.c_str(), "r");
  if (osmFile == NULL)
  {
    cout << "Cant open file: " << fileName << endl;
    exit(1);
  }

  // speedtable and idmap needed
  map<string, int> idMap;
  map<string, size_t> speedTable;
  speedTable["motorway"] = 110;
  speedTable["trunk"] = 110;
  speedTable["primary"] = 70;
  speedTable["secondary"] = 60;
  speedTable["tertiary"] = 50;
  speedTable["motorway_link"] = 50;
  speedTable["trunk_link"] = 50;
  speedTable["primary_link"] = 50;
  speedTable["secondary_link"] = 50;
  speedTable["road"] = 40;
  speedTable["unclassified"] = 40;
  speedTable["residential"] = 30;
  speedTable["unsurfaced"] = 30;
  speedTable["living_street"] = 10;
  speedTable["service"] = 5;

  string nodeKey = "<node";
  string wayKey = "<way";
  size_t found;
  while (fgets(buf, LINE_SIZE, osmFile) != NULL)
  {
    if(_nodes->size() > 0 && _nodes->size() % 1000000 == 0)
      printf("%zu nodes ...\n", _nodes->size());
    string line = buf;
    // Check if the node
    found = line.find(nodeKey);
    if (found != string::npos)
      makeNode(line, &idMap);
    else
    {
      // Check if the way
      found = line.find(wayKey);
      if (found != string::npos)
        makeEdge(line, osmFile, &idMap, speedTable);
    }
  }
}

// _____________________________________________________________________________
void RoadNetwork::getNearestNodesIdsSIMPLE(double latitude, double longitude,
                                      size_t naighboursNumber, double speed,
                                      vector<Edge>* nearestNodes)
{
  // check that values are inside correct ranges...
  assert (-90.0 <= latitude && latitude <= 90.0);
  assert (-180.0 <= longitude && longitude <= 180.0);
  std::priority_queue <std::pair<double, int>,
                  std::vector<std::pair<double, int> >,
                  std::less<std::pair<double, int > > > openNodeQueue;
  for (size_t i = 0; i < (*_nodes).size(); i++)
  {
    double distance = 111.3 * sqrt(
                ((latitude - (*_nodes)[i].latitude) *
               (latitude - (*_nodes)[i].latitude))
              +
               ((longitude - (*_nodes)[i].longitude) *
               (longitude - (*_nodes)[i].longitude))
             );
    if (openNodeQueue.size() == naighboursNumber)
    {
      if (openNodeQueue.top().first > distance)
      {
        openNodeQueue.pop();
        openNodeQueue.push(std::pair<double, int>(distance, i));
      }
    }
    else
      openNodeQueue.push(std::pair<double, int>(distance, i));
  }
  assert(openNodeQueue.size() == naighboursNumber
          || openNodeQueue.size() == _nodes->size());
  while (!openNodeQueue.empty())
  {
    // distance is in km, so we divide it by 5km/h (Wolkingspeed)
    // then multiple the result with 60*60, because we need seconds
    int timeDistance = (openNodeQueue.top().first * 60.0 * 60.0 ) / speed ;
    assert(openNodeQueue.top().second>=0 && timeDistance>=0);
    if (timeDistance < 1)
      timeDistance = 1;
    nearestNodes->push_back(Edge(openNodeQueue.top().second, timeDistance));
    openNodeQueue.pop();
  }
}

// _____________________________________________________________________________
void RoadNetwork::getConnections(int nodeId,  Time arrivalTime, bool isStartNode,
                                 vector<LiveConnection>* outConnections)
{
  vector<Edge>* edges = this->getOutEdges(nodeId);
  // compute all new distances in loop
  int newId;
  int dummyTripId = -1;
  char transfers = 0;
  for (size_t i = 0; i < edges->size(); i++)
  {
    // get all new distances
    Edge& newEdge = (*edges)[i];
    newId = newEdge.otherNodeId;
    int timeNeededForConnection = newEdge.time;
    transfers = 0;
    Connection connection(arrivalTime, arrivalTime, dummyTripId);
    // next arrivaltime should be later
    connection.arrivalTime.addSeconds(timeNeededForConnection);
    LiveConnection newConnection(connection, ROADNETWORK,
                                 newId,
                                 timeNeededForConnection,
                                 transfers, false);
    outConnections->push_back(newConnection);
  }
}


void RoadNetwork::parseFromString(string graphString)
{
  _nodes->clear();
  _outgoingEdges->clear();
  _incomingEdges->clear();
  size_t pos = graphString.find("Nodes:");
  if (pos == string::npos)
  {
    printf("Graphstring kaputt, erwartete \"Nodes:\" am Anfang, statt"
            "dessen %s\n", graphString.c_str());
    return;
  }
  size_t newLinePos = graphString.find('\n');
  if (pos == string::npos)
  {
    printf("Graphstring kaputt, erwarte newline nach nodes-line stattdessen"
            "%s\n", graphString.c_str());
    return;
    
  }
  int nextId;
  double nextLatitude;
  double nextLongitude;
  pos += 7;
  while (pos < newLinePos)
  {
    nextId = atoi(graphString.c_str() + pos);
    if (static_cast<size_t>(nextId) != _nodes->size())
    {
      printf("node id sollte nun %zu sein, "
             "ist aber %d in reststring %s an pos %zu von %s", _nodes->size(),
                                                      nextId,
                                                      graphString.c_str() + pos,
                                                      pos,
                                                      graphString.c_str());
    }
    pos = graphString.find('[', pos) + 1;
    if (pos == string::npos)
    {
      printf("Erwarte Koordinaten für Node in Form [Latitude/Longitude],"
             "finde aber kein \"[\""
              "in %s\n", graphString.c_str());
      return;
    }
    nextLatitude = atof(graphString.c_str() + pos);
    pos = graphString.find('/', pos) + 1;
    if (pos == string::npos)
    {
      printf("Erwarte Koordinaten für Node in Form [Latitude/Longitude],"
             "finde aber kein \"/\""
              "in %s\n", graphString.c_str());
      return;
    }
    nextLongitude = atof(graphString.c_str() + pos);
    pos = graphString.find(']', pos);
    if (pos == string::npos)
    {
      printf("Erwarte Koordinaten für Node in Form [Latitude/Longitude],"
             "finde aber kein \"]\""
              "in %s\n", graphString.c_str());
      return;
    }
    pos = graphString.find(',', pos);
    if (pos != string::npos)
      pos++;
    // HACK: just add the node with its id as lat and long coordinates...
    addNode(NodeExpanded(nextId, nextLatitude, nextLongitude));
  }
  // now parse edges
  pos = graphString.find("Edges:");
  if (pos == string::npos)
  {
    printf("Graphstring kaputt, erwartete \"Edges:\" nach newline, statt"
            "dessen %s\n", graphString.c_str());
    return;
  }
  int startId;
  int endId;
  int time;
  pos = graphString.find('(', pos);
  while (pos != string::npos)
  {
    pos += 1;
    startId = atoi(graphString.c_str() + pos);
    pos = graphString.find("->", pos) + 2;
    endId = atoi(graphString.c_str() + pos);

    pos = graphString.find('|', pos) + 1;
    time = atoi(graphString.c_str() + pos);
    addEdge(startId, endId, time);
    pos = graphString.find('(', pos);
  }
}

string RoadNetwork::toString()
{
  std::stringstream graphStream;

  // add all ndoes, and also add outoging edges...
  graphStream << "Nodes: ";
  for (size_t i = 0; i < this->_nodes->size(); i++)
  {
    if (i > 0) graphStream << ", ";
    NodeExpanded* node = getNode(i);
    graphStream << node->id
                << "[" << node->latitude << "/" << node->longitude << "]";
  }
  graphStream << "\n";
  graphStream << "Edges: ";
  NodeExpanded* curNode;
  NodeExpanded* endNode;
  int time;
  for (size_t i = 0; i < this->_nodes->size(); i++)
  {
    curNode = getNode(i);
    const vector<Edge>& outEdges = _outgoingEdges->at(i);
    for (size_t j = 0; j < outEdges.size(); j++)
    {
      if (i > 0 || j > 0) graphStream << ", ";

      graphStream << "(";
      // first check if corresponding incoming edge exists!
      endNode = getNode(outEdges[j].otherNodeId);
      assert(endNode->id == outEdges[j].otherNodeId);
      bool correspondingIncomingEdgeFound = false;
      const vector<Edge>& inEdges = _incomingEdges->at(endNode->id);
      for (size_t k = 0; k < inEdges.size(); k++)
      {
        if (inEdges[k].otherNodeId == curNode->id)
        {
          correspondingIncomingEdgeFound = true;
          break;
        }
      }
      if (!correspondingIncomingEdgeFound)
      {
        printf("incoming edge fehlt!\n");
        exit(-1);
      }
      time = outEdges[j].time;
      graphStream << curNode->id << " -> ";
      graphStream << endNode->id;
      graphStream << " | " << time;
      graphStream << ")";
    }
  }
  return graphStream.str();
}
