/* Copyright Robin Schirrmeister and Simon Skilevic
 * Lecture Efficient Route Planning summer semester 2011
 * Uni Freiburg
 */

#include "./TransitNetworkExpanded.h"
#include <unordered_map>
#include <unordered_set>
#include <strings.h>
#include <stdio.h>
#include <string>
#include <utility>
#include <vector>
#include <algorithm>
#include <limits>
#include <functional>
#include <sstream>
#include "./CsvParser.h"
#include "./NodeExpanded.h"

using std::string;
using std::unordered_map;
using std::pair;
using std::vector;
using std::unordered_set;

TransitNetworkExpanded::TransitNetworkExpanded()
{
  _verbosity = 1;
}

TransitNetworkExpanded::TransitNetworkExpanded(
                               const TransitNetworkExpanded& orig) : Graph(orig)
{
}

TransitNetworkExpanded::~TransitNetworkExpanded()
{
}

void TransitNetworkExpanded::addEdgeWithTimeDifference(int startId, int endId)
{
  Time startTime = getNode(startId)->time;
  Time endTime = getNode(endId)->time;
  int travelTime = endTime.minusInSeconds(startTime);
  // if endTime is after start time, end time should be interpreted
  // as being on the next day...
  while (travelTime < 0)
    travelTime += 24 *60 * 60;
  assert(travelTime >= 0);
  addEdge(startId, endId, travelTime);
}

void TransitNetworkExpanded::addNodeOfTrip(NodeExpanded node, int tripId)
{
  // check that trip id vector has same size as nodevector
  assert(_tripIds.size() == _nodes->size());
  addNode(node);
  _tripIds.push_back(tripId);
}

void TransitNetworkExpanded::addStop(const Stop& newStop)
{
  assert(newStop.stopId == static_cast<int>(_stops.size()));
  _stops.push_back(newStop);
}

size_t TransitNetworkExpanded::getNumStops()
{
  return _stops.size();
}

bool TransitNetworkExpanded::getStopNodeIds(const string& stopName,
                                          vector<int>* stopNodeIds) const
{
  assert(stopNodeIds->size() == 0);
  // TODO(Robin): instead construct a hashmap
  for (size_t i = 0; i < _stops.size(); i++)
  {
    if (_stops.at(i).stopName == stopName)
    {
      const vector<int>& realStopNodeIds = _stops.at(i).nodeIdsSortedByTime;
      // fill in all stop node ids
      for (size_t j  = 0; j < realStopNodeIds.size(); j++)
      {
        stopNodeIds->push_back(realStopNodeIds[j]);
      }
      return true;
    }
  }
  return false;
}

void TransitNetworkExpanded::getStopsStartingWith(const string& prefix,
                                          vector<string>* stopNames) const
{
  size_t prefixLength = prefix.size();

  for (size_t i = 0; i < _stops.size(); i++)
  {
    string stopName = _stops.at(i).stopName;
    string stopPrefix = stopName.substr(0, prefixLength);
    // compare case insensitive
    if (strcasecmp(stopPrefix.c_str(), prefix.c_str()) == 0)
      stopNames->push_back(stopName);
  }
}

// _____________________________________________________________________________
int TransitNetworkExpanded::getStopId(const string& stopName)
{
  for (size_t stopId = 0; stopId < _stops.size(); stopId++)
  {
    if (_stops.at(stopId).stopName == stopName)
      return stopId;
  }
  return -1;
}

int TransitNetworkExpanded::getTripId(int nodeId)
{
  assert(0 <= nodeId && nodeId < static_cast<int>(_nodes->size()));
  return _tripIds[nodeId];
}

const vector<int>& TransitNetworkExpanded::getStopNodeIds(int stopId)
{
  assert(stopId >= 0 && stopId < static_cast<int>(_stops.size()));
  return _stops.at(stopId).nodeIdsSortedByTime;
}

void TransitNetworkExpanded::getStartNodeIds(int stopId, const Time& startTime,
                                     vector<int>* startNodeIds)
{
  // get the stop node ids sorted by time, and run through them until you find
  // the first node past or equal the start time
  // if it is arrival or transfer node, just return it, otherwise take
  // the first transfer
  // past the starttime
  const vector<int>& nodeIds = getStopNodeIds(stopId);
  if (nodeIds.size() == 0 && _verbosity > 0)
  {
    printf("no nodes at stop: %s(%d)\n", getStopName(stopId).c_str(), stopId);
    return;
  }
  NodeExpanded* node;
  for (size_t i = 0; i < nodeIds.size(); i++)
  {
    node = getNode(nodeIds[i]);
    if (startTime <= node->time)
    {
      if (node->type == DEPARTURENODE)
      {
        startNodeIds->push_back(node->id);
      }
      else if (node->type == TRANSFERNODE)
      {
        startNodeIds->push_back(node->id);
        return;
      }
    }
  }
  // we havent found a start node, so the start should be on the next day
  // now just take the first nodes
  // check if no possible node after the start time found...
  for (size_t i = 0; i < nodeIds.size(); i++)
  {
    node = getNode(nodeIds[i]);
    if (node->type == DEPARTURENODE)
    {
      startNodeIds->push_back(node->id);
    }
    else if (node->type == TRANSFERNODE)
    {
      startNodeIds->push_back(node->id);
      return;
    }
  }
  // should never come here... there should always be a transfer node if there
  // are nodes
  assert(false);
}

// _____________________________________________________________________________
string TransitNetworkExpanded::getStopName(int stopId)
{
  return _stops.at(stopId).stopName;
}
// _____________________________________________________________________________
Stop* TransitNetworkExpanded::getStop(int stopId)
{
  return &(_stops.at(stopId));
}
// _____________________________________________________________________________
void TransitNetworkExpanded::parseFromGTFSFiles(const string& folderName,
                                        char validWeekday,
                                        int transferTimeInSeconds)
{
  _stops.clear();
  // parse which service ids serve which weekdays
  unordered_set<string> validServiceIds;
  if (_verbosity > 1)
    printf("parsing %s ...", (folderName + "/calendar.txt").c_str());
  parseCalendarFile(folderName + "/calendar.txt", validWeekday,
                    &validServiceIds);
  if (_verbosity > 1)
    printf(" done.\n");

  // parse which trips get which trip id...
  unordered_map<string, int> tripIdNameToTripId;
  if (_verbosity > 1)
    printf("parsing %s ...", (folderName + "/trips.txt").c_str());
  parseTripsFile(folderName + "/trips.txt", validServiceIds,
                 &tripIdNameToTripId);
  if (_verbosity > 1)
    printf(" done.\n");

  // parse which stop id name string belongs to which stop id number
  // and to which coordinates
  unordered_map<string, int> stopIdNameToId;
  if (_verbosity > 1)
    printf("parsing %s ...", (folderName + "/stops.txt").c_str());
  vector<Stop> stops;
  parseStopsFile(folderName + "/stops.txt", &stops, &stopIdNameToId);
  if (_verbosity > 1)
    printf(" done.\n");

  // parse which which trip id names have which frequency
  unordered_map<string, vector<TransitParser::TripFrequency> >
                                                              tripIdToFrequency;
  if (_verbosity > 1)
    printf("parsing %s ...", (folderName + "/frequencies.txt").c_str());
  parseFrequenciesFile(folderName + "/frequencies.txt", &tripIdToFrequency);
  if (_verbosity > 1)
    printf(" done.\n");

  // parse stops themselves, create arrival departure and transfer nodes,
  // also consider frequencies..
  if (_verbosity > 1)
    printf("parsing %s ...\n", (folderName + "/stop_times.txt").c_str());
  parseStopTimesFile(folderName + "/stop_times.txt",
                     stopIdNameToId,
                     tripIdToFrequency,
                     tripIdNameToTripId,
                     transferTimeInSeconds);
  if (_verbosity > 1)
  {
    printf("done.\n");
    printf("\nDone parsing.\n");
  }
}

// _____________________________________________________________________________
void TransitNetworkExpanded::parseCalendarFile(const string& calendarFileName,
                                       char validWeekDay,
                                       unordered_set<string>* validServiceIds)
{
  TransitParser parser(_verbosity);
  parser.parseCalendarFile(calendarFileName, validWeekDay, validServiceIds);
}
// _____________________________________________________________________________
void TransitNetworkExpanded::parseTripsFile(const string& tripFileName,
                                   const unordered_set<string>&
                                                          validServiceIds,
                                 unordered_map<string, int>* tripIdNameToTripId)
{
  TransitParser parser(_verbosity);
  parser.parseTripsFile(tripFileName, validServiceIds, tripIdNameToTripId);
}
// _____________________________________________________________________________
void TransitNetworkExpanded::parseStopsFile(const string& stopFileName,
                                    vector<Stop>* stops,
                                    unordered_map<string, int>* stopIdNameToId)
{
  TransitParser parser(_verbosity);
  parser.parseStopsFile(stopFileName, stops, stopIdNameToId);
  // add the stops...
  for (size_t i = 0; i < stops->size(); i++)
  {
    const Stop& nextStop = stops->at(i);
    addStop(nextStop);
  }
}

// _____________________________________________________________________________
void TransitNetworkExpanded::parseFrequenciesFile(
                 const string& frequencyFileName,
                 unordered_map<string, vector<TransitParser::TripFrequency> >*
                  tripIdNameToFrequencies)
{
  TransitParser parser(_verbosity);
  parser.parseFrequenciesFile(frequencyFileName, tripIdNameToFrequencies);
}

// _____________________________________________________________________________
void TransitNetworkExpanded::parseStopTimesFile(const string& stopTimesFileName,
      const unordered_map<string, int>& stopIdNameToId,
      const unordered_map<string, vector<TransitParser::TripFrequency> >&
                                        tripIdNameToFrequencies,
      const unordered_map<string, int>& tripIdNameToTripId,
      int transferTimeInSeconds)
{
  vector<TransitParser::Trip> parsedTrips;
  TransitParser parser(_verbosity);
  parser.parseStopTimesFile(stopTimesFileName, stopIdNameToId,
                            tripIdNameToFrequencies, tripIdNameToTripId,
                            transferTimeInSeconds, &parsedTrips);

  // now add the parsed trips to the network, remember which nodes
  // were added to which stop in order to make
  // transfer edges later
  vector<vector<int> > stopIdsToNodeIds;
  stopIdsToNodeIds.resize(stopIdNameToId.size());
  // add trips
  for (size_t i = 0; i < parsedTrips.size(); i++)
  {
    const TransitParser::Trip& currentTrip = parsedTrips[i];
    // add trip with all freuqncies
    for (size_t j = 0; j < currentTrip.frequencies.size(); j++)
    {
      const TransitParser::TripFrequency& currentFrequency =
                                                    currentTrip.frequencies[j];
      addTrip(currentTrip.nodes, currentFrequency, &stopIdsToNodeIds,
              currentTrip.tripId);
    }
  }
  if (_verbosity > 1)
    printf("  done.\n");
  if (_verbosity > 1)
    printf("  sorting nodes and adding transfer edges...");
  fflush(stdout);
  addTransferEdges(stopIdsToNodeIds);
  if (_verbosity > 1)
    printf("  done.\n");
}

// _____________________________________________________________________________
// TODO(Robin): switch creation of transfer nodes from parsestoptimesfile
// to this function?
void TransitNetworkExpanded::addTrip(
                           const vector<TransitParser::TripNode>& trip,
                           const TransitParser::TripFrequency& frequency,
                           vector<vector<int> >* stopIdToNodeIds,
                           int tripId)
{
  assert(trip[0].secondsSinceDay / (60 * 60) == frequency.startHour);
  assert((trip[0].secondsSinceDay % 3600) / 60 == frequency.startMinute);
  assert(trip[0].secondsSinceDay % 60 == frequency.startSecond);

  // make trip....
  int startSecondInDay = frequency.startHour * 60 * 60 +
                         frequency.startMinute * 60 +
                         frequency.startSecond;
  int endSecondInDay =   frequency.endHour * 60 * 60 +
                         frequency.endMinute * 60 +
                         frequency.endSecond;
  int durationInSeconds = endSecondInDay - startSecondInDay;
  int offsetInSeconds = 0;
  assert(durationInSeconds >= 0);

  int nodeId = this->_nodes->size();

  // always have to finish first trip, this is here
  // in case trip has no real frequency..
  bool tripFinished = false;
  while (!tripFinished)
  {
    for (size_t i = 0; i < trip.size(); i++)
    {
      int newSecondsSinceDay = trip[i].secondsSinceDay + offsetInSeconds;
      // if time is above 24 hours, then it should be adjusted
      while (newSecondsSinceDay >= 24 * 60 * 60)
        newSecondsSinceDay -= 24 * 60 * 60;
      // get latitude and longitude from the stop...
      double latitude = _stops[trip[i].stopId].latitude;
      double longitude = _stops[trip[i].stopId].longitude;
      NodeExpanded nextNode(nodeId, latitude, longitude, trip[i].stopId,
                    Time(newSecondsSinceDay), trip[i].type);
      // remember which weekdays this trip serves...
      addNodeOfTrip(nextNode, tripId);
      assert(_nodes->size() == static_cast<size_t>(nodeId + 1));
      (*stopIdToNodeIds)[nextNode.stopId].push_back(nodeId);

      // now check if we have to add edges
      // first from departure to arrival, here ignore the first arrival node
      // then from arrival to transfer, then from arrival to departure...
      if (nextNode.type == ARRIVALNODE && i > 0)
      {
        assert(i % 3 == 0);
        addEdgeWithTimeDifference(nodeId - 1, nodeId);
      }
      else if (nextNode.type == TRANSFERNODE)
      {
        assert(i % 3 == 1);
        addEdgeWithTimeDifference(nodeId - 1, nodeId);
      }
      else if (nextNode.type == DEPARTURENODE)
      {
        assert(i % 3 == 2);
        addEdgeWithTimeDifference(nodeId - 2, nodeId);
      }
      nodeId++;
    }
    offsetInSeconds += frequency.intervalInSeconds;
    // if no real frequency given, one trip is enough...
    if (frequency.intervalInSeconds == -1)
      tripFinished = true;
    else
      tripFinished =  offsetInSeconds > durationInSeconds;
  }
}

void TransitNetworkExpanded::addTransferEdges(const vector<vector<int> >&
                                         stopIdsToNodeIds)
{
  // Adding the edges from transfer node to next transfer node
  // and all departure nodes inbetween.
  // We do this for each day seperately, sorting the nodes by time
  // and then going through them day by day, also avoiding multiple
  // edges with the same cost inbetween two nodes

  vector<NodeExpanded> nodesThisStop;
  for (size_t stopId = 0; stopId < stopIdsToNodeIds.size(); stopId++)
  {
    const vector<int>& nodeIdsInThisStop = stopIdsToNodeIds[stopId];
    // push back all node pointers for this stop
    nodesThisStop.clear();
    for (size_t i = 0; i < nodeIdsInThisStop.size(); i++)
    {
      NodeExpanded* node = getNode(nodeIdsInThisStop[i]);
      nodesThisStop.push_back(*node);
    }
    // now sort them by time and type
    NodeTimeAndTypeComparator comp;
    sort(nodesThisStop.begin(), nodesThisStop.end(), comp);
    fflush(stdout);

    Stop& currentStop = _stops.at(stopId);
    // now add them to the corresponding stop
    for (size_t i = 0; i < nodesThisStop.size(); i++)
    {
      currentStop.nodeIdsSortedByTime.push_back(nodesThisStop[i].id);
    }
    // now add edges...
    size_t  i = 0;
    // find first transfernode on this day
    while (i < nodesThisStop.size() && nodesThisStop[i].type != TRANSFERNODE)
    {
      i++;
    }
    // now i should be index of first transfer node on the current day..
    // from now on just connect
    // all transfer nodes  until you have gone once through the
    // entire station vector adding edges to departure nodes on the way
    // then also connect last transfer node to first transfer node
    bool reachedEndOfVector = false;
    bool connectedLastToFirstTransferNode = false;
    // check if no trasnfer node found ...
    if (i == nodesThisStop.size())
    {
      reachedEndOfVector = true;
      connectedLastToFirstTransferNode = true;
    }
    while (!reachedEndOfVector || !connectedLastToFirstTransferNode)
    {
      const NodeExpanded& currentNode = nodesThisStop[i];
      assert(currentNode.type == TRANSFERNODE);
      bool foundNextTransferNode = false;
      i++;
      // look for next transfer node and also connect
      // departure nodes  with current transfer node
      while (!foundNextTransferNode)
      {
        if (i == nodesThisStop.size())
        {
          reachedEndOfVector = true;
          i = 0;
        }
        const NodeExpanded& nextNode = nodesThisStop[i];
        if (nextNode.type == TRANSFERNODE)
        {
          addEdgeWithTimeDifference(currentNode.id, nextNode.id);
          foundNextTransferNode = true;
          if (reachedEndOfVector)
            connectedLastToFirstTransferNode = true;
        }
        else if (nextNode.type == DEPARTURENODE)
        {
          addEdgeWithTimeDifference(currentNode.id, nextNode.id);
        }
        if (!foundNextTransferNode) i++;
      }
    }
  }
}

int TransitNetworkExpanded::getVerbosity()
{
  return _verbosity;
}

void TransitNetworkExpanded::setVerbosity(int verbosity)
{
  _verbosity = verbosity;
}

void TransitNetworkExpanded::getConnections(int nodeId, bool isStartNode,
                                         Time arrivalTime,
                                         vector<LiveConnection>* outConnections)
{

    // get outgoing edges from this node
  vector<Edge>* edges = this->getOutEdges(nodeId);
  // compute all new distances in loop
  NodeExpanded* currentNode =  getNode(nodeId);
  int newId;
  int newTripId;
  for (size_t i = 0; i < edges->size(); i++)
  {
    // get all new distances
    Edge& newEdge = (*edges)[i];
    newId = newEdge.otherNodeId;
    newTripId = getTripId(newId);
    NodeExpanded* newNode = getNode(newId);
    int timeNeededForConnection = newEdge.time ;
    char transfers = 0;

    // Time criteria
    // Transfer criteria
    if ((currentNode->type == TRANSFERNODE
     && newNode->type == DEPARTURENODE)
     && !isStartNode)
      transfers = 1;
    Connection connection(arrivalTime, arrivalTime, newTripId);
    connection.arrivalTime.addSeconds(timeNeededForConnection);
    LiveConnection newConnection(connection, TRANSITNETWORK,
                                 newId,
                                 timeNeededForConnection,
                                 transfers, false);
    outConnections->push_back(newConnection);
  }
}