#include <getopt.h>
#include <iostream>
#include <string>
#include <vector>
#include "./EdgeDependent.h"
#include "./TransitNetworkDependent.h"
#include "./Dijkstra.h"
#include "./ExpandedDijkstra.h"
#include "./DependentDijkstra.h"
#include "./MulticriteriaDependentDijkstra.h"
#include "./Timer.h"
#include "./MulticriteriaExpandedDijkstra.h"

using std::string;
using std::vector;
using std::cout;
using std::endl;

unsigned int randomSeedPerf;
// _____________________________________________________________________________
void printUsage()
{
  printf("This program measures performance for shortest path queries "
         "transit networks.\n\n"
         "USAGE: ./TransitNetworkDependentPerf [OPTIONS] GTFSFolder\n"
         "OPTIONS:"
          "--(h)elp: Display this usage info\n"
          "--(q)ueries NUMBER: make NUMBER many queries\n");
}
// _____________________________________________________________________________
int getRandomNodeId(int size)
{
  int idToTry = rand_r(&randomSeedPerf) % size;
  return idToTry;
}
// _____________________________________________________________________________
int main(int argc, char** argv)
{
  string GTFSFolder;
  // Needed for getopt_long.
  int optChr;
  int numberOfQueries = 100;
  optind  = 1;
  while (true)
  {
    int optionIndex = 0;
    // Define command-line options.
    static struct option longOptions[] =
    {
      {"help", no_argument, 0, 'h'},
      {"queries", no_argument, 0, 'q'},
      {0, 0, 0, 0}
    };

    optChr = getopt_long(argc, argv, "hq:", longOptions, &optionIndex);

    if (optChr == -1) break;

    switch (optChr)
    {
      case 0:
        break;
      case 'h':
        printUsage();
        exit(0);
      case 'q':
        numberOfQueries = atoi(optarg);
        break;
      default:
        printf("unknown option: %c\n", optChr);
        printUsage();
        return 1;
        break;
    }
  }

  // check that GTFS folder was given
  if (argc - optind != 1)
  {
    printf("Please specify GTFS input folder!\n\n");
    printUsage();
    return 1;
  }
  // at the moment using monday...
  char testDay = 0;
  int transferTime = 4 * 60;
  GTFSFolder = argv[optind];
  // parse transit network
  TransitNetworkDependent transitNetworkDependent;
  TransitNetworkExpanded transitNetworkExpanded;
  cout << "Parsing GTFS File by dependent network:" << endl;
  transitNetworkDependent.parseFromGTFSFiles(GTFSFolder);
  cout << " done." << endl;
  cout << "Parsing GTFS File by expanded network:" << endl;
  transitNetworkExpanded.parseFromGTFSFiles(GTFSFolder, testDay, transferTime);
  cout << " done." << endl;
  DependentDijkstra routePlannerDependent(&transitNetworkDependent);
  MulticriteriaDependentDijkstra routePlannerMulticriteriaDependent(
                                                      &transitNetworkDependent);
  MulticriteriaExpandedDijkstra routePlannerMulticriteriaExpanded(
                                                       &transitNetworkExpanded);
  ExpandedDijkstra routePlannerExpanded(&transitNetworkExpanded);
  Timer timer;
  size_t timeSum = 0;
  vector< vector <int> > randomInput;
  randomInput.resize(numberOfQueries);
  cout << "Creating of test data:" << endl;
  for (int i = 0; i < numberOfQueries; i++)
  {
    int startId = getRandomNodeId(transitNetworkDependent.getNumNodes());

    int endId = startId;
    while (startId == endId)
      endId = getRandomNodeId(transitNetworkDependent.getNumNodes());
    int hour = getRandomNodeId(24);
    int min = getRandomNodeId(60);
    int sec = getRandomNodeId(60);
    randomInput[i].push_back(startId); // 0
    randomInput[i].push_back(endId);   // 1
    randomInput[i].push_back(hour);    // 2
    randomInput[i].push_back(min);     // 3
    randomInput[i].push_back(sec);     // 4
  }
  timeSum = 0;
  cout << endl << "done." << endl;
  cout << "Testing of dependent Network:" << endl;
  for (int i = 0; i < numberOfQueries; i++)
  {
    Time startTime(randomInput[i][2], randomInput[i][3], randomInput[i][4]);
    timer.start();
    routePlannerDependent.computeShortestPath(randomInput[i][0], startTime,
                                              randomInput[i][1], false);
    timer.stop();
    timeSum += timer.usecs();
  }
  printf("Time taken for dependent Dijkstra: %7.2f ms\n",timeSum / 1000.0);
  printf("Average Time: %7.2f ms\n", timeSum / (1000.0 * numberOfQueries));
  cout << "Testing of expanded Network:" << endl;
  timeSum = 0;
  for (int i = 0; i < numberOfQueries; i++)
  {
    int startStopId = randomInput[i][0];
    int endStopId = randomInput[i][1];
    Time startTime(randomInput[i][2], randomInput[i][3], randomInput[i][4]);
    timer.start();
    routePlannerExpanded.computeShortestPath(startStopId, startTime,
                                             endStopId, false);
    timer.stop();
    timeSum += timer.usecs();
  }
  printf("Time taken for expanded Dijkstra simple: %7.2f ms\n",
                                               timeSum / 1000.0);
  printf("Average Time: %7.2f ms\n",
          timeSum / (1000.0 * numberOfQueries));

  cout << "Testing of dependentMulticriteria Network:" << endl;
  timeSum = 0;
  for (int i = 0; i < numberOfQueries; i++)
  {
    int startStopId = randomInput[i][0];
    int endStopId = randomInput[i][1];
    Time startTime(randomInput[i][2], randomInput[i][3], randomInput[i][4]);
    timer.start();
    routePlannerDependent.computeShortestPath(startStopId, startTime, endStopId, false);
    routePlannerMulticriteriaDependent.transferMax = routePlannerDependent.transfers;
    routePlannerMulticriteriaDependent.timeMax = routePlannerDependent.time;
    routePlannerMulticriteriaDependent.computeShortestPath(startStopId,
                                             startTime,
                                             endStopId, false);
    timer.stop();
    timeSum += timer.usecs();
  }
  printf("Time taken for dependent multicriteria Dijkstra : %7.2f ms\n",
                                               timeSum / 1000.0);
  printf("Average Time: %7.2f ms\n",
          timeSum / (1000.0 * numberOfQueries));

  cout << "Testing of multicriteria expanded Network:" << endl;
  timeSum = 0;
  for (int i = 0; i < numberOfQueries; i++)
  {
    int startStopId = randomInput[i][0];
    int endStopId = randomInput[i][1];
    Time startTime(randomInput[i][2], randomInput[i][3], randomInput[i][4]);
    timer.start();
    if (!(routePlannerDependent.computeShortestPath(startStopId, startTime,
                                              endStopId, false) < 0))
    routePlannerMulticriteriaExpanded.computeShortestPath(startStopId,
                                                          startTime,
                                                          endStopId,
                                                          false);
    timer.stop();
    timeSum += timer.usecs();
  }
  printf("Time taken for expanded multicriteria Dijkstra simple: %7.2f ms\n",
                                               timeSum / 1000.0);
  printf("Average Time: %7.2f ms\n",
          timeSum / (1000.0 * numberOfQueries));
  return 0;
}
