// Copyright November 8, 2011 Robin Schirrmeister and Simon Skilevic
// Chair of Algorithms and Datastructures Uni Freiburg


#include "./TransitPath.h"
#include <assert.h>
#include <unordered_set>
#include <algorithm>
#include <vector>
#include <utility>
#include <string>
#include <sstream>
#include "./EdgeDependent.h"
#include "./LiveConnection.h"

using std::vector;
using std::unordered_set;
using std::stringstream;

TransitPath::TransitPath(const Time& startTime)
{
  _startTime = startTime;
}
void TransitPath::clear()
{
  _pathNodes.clear();
}

void TransitPath::reverse()
{
  std::reverse(_pathNodes.begin(), _pathNodes.end());
}

size_t TransitPath::size() const
{
  return _pathNodes.size();
}

bool TransitPath::isEqualTo(const TransitPath& otherPath) const
{
  if (!(this->_startTime == otherPath._startTime))
    return false;
  if (this->_pathNodes.size() != otherPath._pathNodes.size())
    return false;
  for (size_t i = 0; i < this->_pathNodes.size(); i++)
  {
    if (!this->_pathNodes[i].isEqualTo(otherPath._pathNodes[i]))
      return false;
  }
  return true;
}

bool TransitPath::hasSameResult (const TransitPath& otherPath) const
{
  return this->getDurationInMinutes() == otherPath.getDurationInMinutes() &&
         this->getTransfers() == otherPath.getTransfers();
}

void TransitPath::addStation(PathNodeType type, int stationId,
                             double latitude, double longitude,
                             Time arrivalTime, Time departureTime,
                             int arrivedWithTripId)
{
  _pathNodes.push_back(PathNode(type, stationId, latitude, longitude,
                       arrivalTime, departureTime, arrivedWithTripId));
}

TransitPath::PathNodeType TransitPath::getType(int pathNodeIndex)
{
  assert(0 <= pathNodeIndex &&
         pathNodeIndex < static_cast<int>(_pathNodes.size()));
  return _pathNodes[pathNodeIndex].type;
}
int TransitPath::getStationId(int pathNodeIndex)
{
  assert(0 <= pathNodeIndex &&
         pathNodeIndex < static_cast<int>(_pathNodes.size()));
  return _pathNodes[pathNodeIndex].stationId;
}

double TransitPath::getLatitude(int pathNodeIndex)
{
  assert(0 <= pathNodeIndex &&
         pathNodeIndex < static_cast<int>(_pathNodes.size()));
  return _pathNodes[pathNodeIndex].latitude;
  
}

double TransitPath::getLongitude(int pathNodeIndex)
{
  assert(0 <= pathNodeIndex &&
         pathNodeIndex < static_cast<int>(_pathNodes.size()));
  return _pathNodes[pathNodeIndex].longitude;

}

const Time& TransitPath::getArrivalTime(int pathNodeIndex)
{
  assert(0 <= pathNodeIndex &&
         pathNodeIndex < static_cast<int>(_pathNodes.size()));
  return _pathNodes[pathNodeIndex].arrivalTime;
}

const Time& TransitPath::getDepartureTime(int pathNodeIndex)
{
  assert(0 <= pathNodeIndex &&
         pathNodeIndex < static_cast<int>(_pathNodes.size()));
  return _pathNodes[pathNodeIndex].departureTime;
}
int TransitPath::getArrivedWithTripId(int pathNodeIndex)
{
  assert(0 <= pathNodeIndex &&
         pathNodeIndex < static_cast<int>(_pathNodes.size()));
  return _pathNodes[pathNodeIndex].arrivedWithTripId;
}
const Time& TransitPath::getStartTime() const
{
  return _startTime;
}

// get the amount of transfers on that trip
int TransitPath::getTransfers() const
{
  // now transfers defined as driving in a "new" train, i.e.
  // either changing trains or moving from road network to transit network
  // AND taking a train in the transit network(i.e. not just moving
  // to the end station in the transit network)
  // need atleast 3 nodes for a transfer...
  if (_pathNodes.size() <= 2) return 0;
  int transfers = 0;
  // trip id of pathnode nr. 0 is always -1 therefore do not count it
  int lastTripId = _pathNodes[1].arrivedWithTripId;
  int curTripId;
  PathNodeType curType;
  for (size_t i = 2; i < _pathNodes.size(); i++)
  {
    curTripId = _pathNodes[i].arrivedWithTripId;
    curType = _pathNodes[i].type;
    if (curTripId != lastTripId && curType == TRANSITNODE)
      transfers++;
    lastTripId = curTripId;
  }
  return transfers;
}
// get the duration of that trip!
double TransitPath::getDurationInMinutes() const
{
  if (_pathNodes.size() == 0) return 0;
  // path with one ndoe does not make sense...
  assert(_pathNodes.size() != 1);
  double duration = 0;
  // go through trip, count duration, respect dayboundaries
  size_t i = 0;
  Time oldArrivalTime = _pathNodes[0].arrivalTime;
  Time nextArrivalTime;
  // TODO(Robin): instead just go and always take time between each node?
  // maybe not even slower?
  while (i + 1 < _pathNodes.size())
  {
    Time curArrivalTime = _pathNodes[i].arrivalTime;
    nextArrivalTime = _pathNodes[i + 1].arrivalTime;
    if (nextArrivalTime < curArrivalTime)  // crossing a day boundary
    {
      // take time the trip needed in the "old" day plus the
      // time taken at the change from old day to new day and continue counting
      duration += curArrivalTime.minusInSeconds(oldArrivalTime);
      duration += nextArrivalTime.minusInSecondsDependent(curArrivalTime);
      oldArrivalTime = nextArrivalTime;
    }
    i++;
  }
  duration += nextArrivalTime.minusInSeconds(oldArrivalTime);
  return duration / 60.0;
}

string TransitPath::toString() const
{
  Time dummyTime;
  TransitPath dummyPath(dummyTime);
  return createDiffOrString(dummyPath, false);
}
string TransitPath::createJSON() const
{
  stringstream jsonStringStream;
  // add duration, transfers and nodes
  jsonStringStream << "{\n"
                   << "\"startTime\" :" << _startTime.createJSON() << ",\n"
                   << "\"duration\": " << getDurationInMinutes() << ",\n"
                   << "\"transfers\": " << getTransfers() << ",\n"
                   << "\"nodes\": " << "[\n";
  for (size_t i = 0; i < _pathNodes.size(); i++)
  {
    if (i > 0) jsonStringStream << ",";
    jsonStringStream << _pathNodes[i].createJSON();
  }
  jsonStringStream <<                 "]\n"
                   << "}\n";
  return jsonStringStream.str();
}
pair<vector<int>, vector<int> > TransitPath::findDifferentPaths(
                                         const vector<TransitPath>& firstPaths,
                                         const vector<TransitPath>& secondPaths)
{
  pair<vector<int>, vector<int> > result;
  unordered_set<int> foundPartnersInSecondPath;
  // first go through the first paths, find out which ones lack partners,
  // already remember those of second paths that have partners,
  // then check which ones from second paths havent found a partner
  for (size_t i = 0; i < firstPaths.size(); i++)
  {
    bool foundPartner = false;
    const TransitPath& firstPath = firstPaths[i];
    for (size_t j = 0; j < secondPaths.size(); j++)
    {
      const TransitPath& secondPath = secondPaths[j];
      if (firstPath.isEqualTo(secondPath))
      {
        foundPartner = true;
        foundPartnersInSecondPath.insert(j);
        break;
      }
    }
    if (!foundPartner)
      result.first.push_back(i);
  }
  // check which ones from second paths ahvent found a partner
  for (size_t i = 0; i < secondPaths.size(); i++)
  {
    if (foundPartnersInSecondPath.count(i) == 0)
      result.second.push_back(i);
  }
  return result;
}

pair<vector<int>, vector<int> > TransitPath::findDifferentResults(
                                        const vector<TransitPath>& firstPaths,
                                        const vector<TransitPath>& secondPaths)
{
 pair<vector<int>, vector<int> > result;
  unordered_set<int> foundPartnersInSecondPath;
  // first go through the first paths, find out which ones lack partners,
  // already remember those of second paths that have partners,
  // then check which ones from second paths havent found a partner
  for (size_t i = 0; i < firstPaths.size(); i++)
  {
    bool foundPartner = false;
    const TransitPath& firstPath = firstPaths[i];
    for (size_t j = 0; j < secondPaths.size(); j++)
    {
      const TransitPath& secondPath = secondPaths[j];
      if (firstPath.hasSameResult(secondPath))
      {
        foundPartner = true;
        foundPartnersInSecondPath.insert(j);
        break;
      }
    }
    if (!foundPartner)
      result.first.push_back(i);
  }
  // check which ones from second paths ahvent found a partner
  for (size_t i = 0; i < secondPaths.size(); i++)
  {
    if (foundPartnersInSecondPath.count(i) == 0)
      result.second.push_back(i);
  }
  return result;
}
string TransitPath::createDiffTo(const TransitPath& otherPath) const
{
  return createDiffOrString(otherPath, true);
}
string TransitPath::createDiffOrString(const TransitPath& otherPath,
                                      bool createDiff) const
{
  string resultDiff;
  const size_t MAXBUFFERLENGTH = 1000;
  char diffBuffer[MAXBUFFERLENGTH + 1];
  if (this->_startTime == otherPath._startTime || !createDiff)
  {
    snprintf(diffBuffer, MAXBUFFERLENGTH, "Starttime: %s\n",
            this->_startTime.toString().c_str());
  }
  else
  {
    snprintf(diffBuffer, MAXBUFFERLENGTH, "Starttimes: %s | %s\n",
                                       this->_startTime.toString().c_str(),
                                       otherPath._startTime.toString().c_str());
  }
  resultDiff.append(diffBuffer);
  if (this->getDurationInMinutes() == otherPath.getDurationInMinutes() ||
      !createDiff)
  {
    snprintf(diffBuffer, MAXBUFFERLENGTH, "Duration: %7.5f\n",
            this->getDurationInMinutes());
  }
  else
  {
    snprintf(diffBuffer, MAXBUFFERLENGTH, "Durations: %13.9f | %13.9f\n",
                                       this->getDurationInMinutes(),
                                       otherPath.getDurationInMinutes());
  }
  resultDiff.append(diffBuffer);
  if (this->getTransfers() == otherPath.getTransfers() || !createDiff)
  {
    snprintf(diffBuffer, MAXBUFFERLENGTH, "Transfers: %2d\n",
            this->getTransfers());
  }
  else
  {
    snprintf(diffBuffer, MAXBUFFERLENGTH, "Transfers: %2d | %2d\n",
                                       this->getTransfers(),
                                       otherPath.getTransfers());
  }
  resultDiff.append(diffBuffer);
  // print trip diff until only one trip has more remaining nodes
  size_t i = 0;
  for (; i < std::min(this->_pathNodes.size(), otherPath._pathNodes.size());
       i++)
  {
    const PathNode& thisNode = this->_pathNodes[i];
    const PathNode& otherNode = otherPath._pathNodes[i];
    if (thisNode.isEqualTo(otherNode))
    {
      snprintf(diffBuffer, MAXBUFFERLENGTH, "%-3zu %-7s "
              "%5d ARR:%s   DEP:%s Trip: %4d\n",
              i,
              thisNode.type == ROADNODE ? "Road" : "Transit",
              thisNode.stationId, thisNode.arrivalTime.toString().c_str(),
              thisNode.departureTime.toString().c_str(),
              thisNode.arrivedWithTripId);
    }
    else
    {
      snprintf(diffBuffer, MAXBUFFERLENGTH, "%-3zu %-7s "
        "%5d ARR:%s   DEP:%s Trip: %4d --- "
        "%-7s %5d ARR:%s   DEP:%s Trip: %4d \n",
              i,
              thisNode.type == ROADNODE ? "Road" : "Transit",
              thisNode.stationId, thisNode.arrivalTime.toString().c_str(),
              thisNode.departureTime.toString().c_str(),
              thisNode.arrivedWithTripId,
              otherNode.type == ROADNODE ? "Road" : "Transit",
              otherNode.stationId, otherNode.arrivalTime.toString().c_str(),
              otherNode.departureTime.toString().c_str(),
              otherNode.arrivedWithTripId);
    }
    resultDiff.append(diffBuffer);
  }
  // print the reamining trip nodes (only one path has more nodes from here on
  for (; i < this->_pathNodes.size(); i++)
  {
    const PathNode& thisNode = this->_pathNodes[i];
      snprintf(diffBuffer, MAXBUFFERLENGTH, "%-3zu %-7s "
              "%5d ARR:%s   DEP:%s Trip: %4d\n",
              i,
              thisNode.type == ROADNODE ? "Road" : "Transit",
              thisNode.stationId, thisNode.arrivalTime.toString().c_str(),
              thisNode.departureTime.toString().c_str(),
              thisNode.arrivedWithTripId);
    resultDiff.append(diffBuffer);
  }
  for (; i < otherPath._pathNodes.size(); i++)
  {
    const PathNode& otherNode = otherPath._pathNodes[i];
    snprintf(diffBuffer, MAXBUFFERLENGTH,
            "%-3zu %64s %2d ARR:%s   DEP:%s Trip: %4d\n",
            i,
            otherNode.type == ROADNODE ? "Road" : "Transit",
            otherNode.stationId, otherNode.arrivalTime.toString().c_str(),
            otherNode.departureTime.toString().c_str(),
            otherNode.arrivedWithTripId);
    resultDiff.append(diffBuffer);
  }
  return resultDiff;
}

TransitPath::PathNode::PathNode(PathNodeType type, int stationId,
                                double latitude, double longitude,
                                Time arrivalTime, Time departureTime,
                                int arrivedWithTripId)
{
  this->type = type;
  this->stationId = stationId;
  this->latitude = latitude;
  this->longitude = longitude;
  this->arrivalTime = arrivalTime;
  this->departureTime = departureTime;
  this->arrivedWithTripId = arrivedWithTripId;
}

bool TransitPath::PathNode::isEqualTo (
                                    const TransitPath::PathNode& otherPathNode)
                                    const
{
  if (this->type == otherPathNode.type &&
      this->stationId == otherPathNode.stationId &&
      this->latitude == otherPathNode.latitude &&
      this->longitude == otherPathNode.longitude &&
      this->arrivalTime == otherPathNode.arrivalTime &&
      this->departureTime == otherPathNode.departureTime &&
      this->arrivedWithTripId == otherPathNode.arrivedWithTripId)
    return true;
  else
    return false;
}

string TransitPath::PathNode::createJSON() const
{
  stringstream jsonStream;
  jsonStream << "{\n";
  jsonStream << "  \"type\": ";
  if (this->type == ROADNODE)
    jsonStream << 0;
  else if (this->type == TRANSITNODE)
    jsonStream << 1;
  else
    jsonStream << "\"UNKNOWN\"";
  jsonStream << ",\n"
             << "  \"locationId\": " << this->stationId << ",\n"
             << "  \"latitude\": " << this->latitude  << ",\n"
             << "  \"longitude\": " << this->longitude  << ",\n"
             << "  \"arrivalTime\": " << this->arrivalTime.createJSON()
             << ",\n"
             << "  \"departureTime\": " << this->departureTime.createJSON()
             << ",\n"
             << "  \"arrivedWithTripId\": " << this->arrivedWithTripId
             << "\n"
             << "}\n";
  return jsonStream.str();
}
