// Copyright 2011 Simon Skilevic und Robin Schirrmeister

#include <getopt.h>
#include <boost/asio.hpp>
#include <algorithm>
#include <ctime>
#include <iostream>
#include <string>
#include <vector>
#include <utility>
#include "./Heuristics.h"
#include "./TransitNetworkExpanded.h"
#include "./ExpandedDijkstra.h"
#include "./Timer.h"

using boost::asio::ip::tcp;
using std::vector;
int port = 5273;

// For formatting ASCII output.
#define BOLD     "\033[1m"
#define BLACK    "\033[30m"
#define RED      "\033[31m"
#define GREEN    "\033[32m"
#define BROWN    "\033[33m"
#define BLUE     "\033[34m"
#define RESET    "\033[0m"

void printUsage()
{
  printf("This program starts a server for a transit network.\n\n"
         "USAGE: ./TransitNetworkServerMain GTFSFolder\n");
}

void replaceInString(string* original, const string& toReplace,
                       const string& replaceWith)
{
  size_t pos = original->find(toReplace);
  while (pos != string::npos)
  {
    original->replace(pos, toReplace.size(), replaceWith);
    pos = original->find(toReplace, pos + 1);
  }
}

using std::string;
using std::vector;
int main(int argc, char** argv)
{
  string GTFSFolder;
  // Needed for getopt_long.
  int optChr;
  optind  = 1;
  while (true)
  {
    int optionIndex = 0;
    // Define command-line options.
    static struct option longOptions[] =
    {
      {"help", no_argument, 0, 'h'},
      {0, 0, 0, 0}
    };

    optChr = getopt_long(argc, argv, "h", longOptions, &optionIndex);

    if (optChr == -1) break;

    switch (optChr)
    {
      case 0:
        break;
      case 'h':
        printUsage();
        exit(0);
      default:
        printf("unknown option: %c\n", optChr);
        printUsage();
        return 1;
        break;
    }
  }

  // check that GTFS folder was given
  if (argc - optind != 1)
  {
    printf("Please specify GTFS input folder!\n\n");
    printUsage();
    return 1;
  }
  GTFSFolder = argv[optind];
  // parse transit network for one day
  char validWeekday = 0;
  int transferTimeInSeconds = 4 * 60;
  TransitNetworkExpanded transitNetwork;
  transitNetwork.parseFromGTFSFiles(GTFSFolder, validWeekday,
                                    transferTimeInSeconds);
  ExpandedDijkstra routePlanner(&transitNetwork);
  printf("Nodes in Transit Network: %zu\n", transitNetwork.getNumNodes());

  // Now the server part.
  try
  {
    // Create an boost IO service.
    boost::asio::io_service io_service;

    // Create socket and bind to and listen on given port.
    tcp::acceptor acceptor(io_service, tcp::endpoint(tcp::v4(), port));

    // Wait for requests and process them.
    while (true)
    {
      // Wait for request.
      std::cout << BLUE << BOLD << "Waiting for query on port " << port
                << " ... " << RESET << std::flush;
      tcp::socket socket(io_service);
      acceptor.accept(socket);
      std::time_t now = std::time(0);
      std::string daytime = std::ctime(&now); // NOLINT
      std::cout << "done, received new request on " << daytime << std::flush;

      // Get the request string.
      std::vector<char> requestBuffer(1000);
      boost::system::error_code read_error;
      socket.read_some(boost::asio::buffer(requestBuffer), read_error);
      std::string request(requestBuffer.size(), 0);
      std::copy(requestBuffer.begin(), requestBuffer.end(), request.begin());
      for (size_t i = 0; i < request.size(); i++)
        request[i] = isspace(request[i]) ? ' ' : request[i];
      std::cout << "request string is \"" << request << std::endl;

      // Extract query = the thing between the GET and the HTTP ...

      std::stringstream jsonp;  // the answer jsonp text...
      std::string query = request.substr(5);
      size_t pos = query.find(" HTTP");
      if (pos != std::string::npos) query.erase(pos);
      std::cout << "query: " << query << "\n";
      string queryCommand = "NEWQUERY:";
      string stationCommand = "GETSTATIONNAMES:";
      if (query.find(queryCommand) != string::npos)
      {
        // new query..
        query = query.substr(queryCommand.size());
        std::cout << "query: " << query << "\n";
        size_t posStartStation = query.find("||");
        string startStation = query.substr(0, posStartStation);
        replaceInString(&startStation, string("%20"), " ");
        std::cout << "start station: " << startStation << "\n";
        size_t posEndStation = query.find("||", posStartStation + 1);
        string endStation = query.substr(posStartStation + 2,
                                         posEndStation - (posStartStation + 2));
        replaceInString(&endStation, string("%20"), " ");
        std::cout << "end station: " << endStation << "\n";
        int day = atoi(query.c_str() + posEndStation + 2);
        size_t hourPos = query.find("||", posEndStation + 1);
        int hour = atoi(query.c_str() + hourPos + 2);
        size_t minutePos = query.find("||", hourPos + 1);
        int minute = atoi(query.c_str() + minutePos + 2);
        std::cout << "day " << day << "\n";
        std::cout << "hour" << hour << "\n";
        std::cout << "minute" << minute << "\n";
        Time startTime(hour, minute, 0);
        routePlanner.computeShortestPath(transitNetwork.getStopId(startStation),
                                         startTime,
                                         transitNetwork.getStopId(endStation),
                                         true);
        const vector<pair<NodeExpanded, Edge> >& path =
                                               routePlanner.transitShortestPath;
        jsonp << "handleNewResult(\n"
              << "{\n"
              << " \"path\" :  [\n";
        // enter path as sequence of nodes for now
        vector<NodeExpanded> clearPath;
        for (size_t i = 0; i < path.size(); i++)
        {
          const NodeExpanded& currentNode = path.at(i).first;
          if (currentNode.type == DEPARTURENODE ||
              (i > 0  && currentNode.type == ARRIVALNODE))
          {
            clearPath.push_back(currentNode);
          }
        }

        for (size_t i = 0; i < clearPath.size(); i++)
        {
          NodeExpanded& currentNode = clearPath[i];
          jsonp << "    {\n";
          string stopName = transitNetwork.getStopName(currentNode.stopId);
          jsonp << "\"stop\": \"" << stopName << "\",\n";
          jsonp << "\"hour\":"  << static_cast<int>(currentNode.time.hour())
                << ",\n";
          jsonp << "\"minute\":" << static_cast<int>(currentNode.time.minute())
                << "\n";
          jsonp << "    }";
          if (i < clearPath.size() - 1)
            jsonp << ",";
          jsonp << "\n";
        }
        jsonp << "  ]\n";
        jsonp << "})\n";
      }
      else if (query.find(stationCommand) != string::npos)
      {
        jsonp << "handleStationNames(\n"
              << "{ \"stopNames\" : [";
        // get all stop names
        vector<string> stopNames;
        for (size_t i = 0; i < transitNetwork.getNumStops(); i++)
          stopNames.push_back(transitNetwork.getStopName(i));
        for (size_t i = 0; i < stopNames.size(); i++)
        {
          jsonp << "\"" << stopNames[i] << "\"";
          if (i < stopNames.size() - 1)
            jsonp << ",";
        }
        jsonp << "]})\n";
      }
      std::ostringstream answer;
      answer << "HTTP/1.0 200 OK\r\n"
             << "Content-Length: " << jsonp.str().size() << "\r\n"
             << "Content-Type: application/javascript" << "\r\n"
             << "Connection: close\r\n"
             << "\r\n"
             << jsonp.str();

      printf("\nAnswertext:  %s\n", answer.str().c_str());
      boost::system::error_code write_error;
      boost::asio::write(socket, boost::asio::buffer(answer.str()),
          boost::asio::transfer_all(), write_error);
    }
  }
  catch (std::exception& e) // NOLINT
  {
    std::cerr << e.what() << std::endl;
  }
}

