// Copyright 2011 Simon Skilevic und Robin Schirrmeister

#include <boost/asio.hpp>
#include <algorithm>
#include <ctime>
#include <iostream>
#include <string>
#include <vector>
#include "./RoadNetwork.h"
#include "./Dijkstra.h"
#include "./Heuristics.h"

using boost::asio::ip::tcp;
using std::vector;
int port = 5274;

// For formatting ASCII output.
#define BOLD     "\033[1m"
#define BLACK    "\033[30m"
#define RED      "\033[31m"
#define GREEN    "\033[32m"
#define BROWN    "\033[33m"
#define BLUE     "\033[34m"
#define RESET    "\033[0m"

// Main program implementing a simple server loop using boost:asio for the
// socket communication. Expects a simple HTTP get request and returns a json
// object containing the start and endnode and the path from start to endnode
// the path is simply an array of coordinates of the nodes on the path, i.e.:
// latitude 1, longitude 1, latitude 2, longitude 2, ...

int main(int argc, char** argv)
{
  // Parse command line arguments.
  if (argc != 2)
  {
    std::cout << "Usage: ./mapServerMain <osm file name>" << std::endl
         << "Start a server answering closest point queries" << std::endl;
    exit(1);
  }
  std::string osmFileName = argv[1];

  // Read road network and create contracted graph.
  Graph roadNetwork;
  roadNetwork.readFromOsmFile(osmFileName);
  roadNetwork.reduceToBiggestComponent();
  NoHeuristic noHeuristic;
  StraightLineHeuristic aStarHeuristic;
  SimpleDijkstra routePlanner(&roadNetwork, &aStarHeuristic);
  // roadNetwork.createEdgeFlags(100);
  // roadNetwork.createContractedGraph();

  // Now the server part.
  try
  {
    // Create an boost IO service.
    boost::asio::io_service io_service;

    // Create socket and bind to and listen on given port.
    tcp::acceptor acceptor(io_service, tcp::endpoint(tcp::v4(), port));

    // Wait for requests and process them.
    while (true)
    {
      // Wait for request.
      std::cout << BLUE << BOLD << "Waiting for query on port " << port
                << " ... " << RESET << std::flush;
      tcp::socket socket(io_service);
      acceptor.accept(socket);
      std::time_t now = std::time(0);
      std::string daytime = std::ctime(&now); // NOLINT
      std::cout << "done, received new request on " << daytime << std::flush;

      // Get the request string.
      std::vector<char> requestBuffer(1000);
      boost::system::error_code read_error;
      socket.read_some(boost::asio::buffer(requestBuffer), read_error);
      std::string request(requestBuffer.size(), 0);
      std::copy(requestBuffer.begin(), requestBuffer.end(), request.begin());
      for (size_t i = 0; i < request.size(); i++)
        request[i] = isspace(request[i]) ? ' ' : request[i];
      std::cout << "request string is \"" << request << std::endl;

      // Extract query = the thing between the GET and the HTTP ...
      std::string query = request.substr(5);
      size_t pos = query.find(" HTTP");
      if (pos != std::string::npos) query.erase(pos);

      // Extract latitude and longitude of both nodes.
      double startLatitude = 0.0;
      double startLongitude = 0.0;
      double endLatitude = 0.0;
      double endLongitude = 0.0;
      pos = query.find('(');
      if (pos != std::string::npos)
        startLatitude = atof(query.substr(pos + 1).c_str());
      pos = query.find(",%20", pos + 1);
      if (pos != std::string::npos)
        startLongitude = atof(query.substr(pos + 4).c_str());

      pos = query.find('(', pos + 1);
      if (pos != std::string::npos)
        endLatitude = atof(query.substr(pos + 1).c_str());
      pos = query.find(",%20", pos + 1);
      if (pos != std::string::npos)
        endLongitude = atof(query.substr(pos + 4).c_str());

      // Find closest OSM node.
      int closestStartNodeId = roadNetwork.findIdOfClosestNode(startLatitude,
                                                                startLongitude);
      int closestEndNodeId   = roadNetwork.findIdOfClosestNode(endLatitude,
                                                                endLongitude);
      NodeExpanded* closestStartNode;
      NodeExpanded* closestEndNode;
      closestStartNode = roadNetwork.getNode(closestStartNodeId);
      closestEndNode = roadNetwork.getNode(closestEndNodeId);

      routePlanner.computeShortestPath(closestStartNode->id,
                                       closestEndNode->id,
                                       true, true);
      vector<NodeExpanded>* shortestPath = routePlanner.shortestPath;
      // Send a JSON object containing the latitude and longitude from
      // the startnode and the endnode as well as the path from start to endnode
      std::stringstream jsonp;
      jsonp << "callback(\n"
            << "{\n"
            << "  \"startNode\" : \n"
            << "  {\n"
            << "    \"latitude\" : "  << closestStartNode->latitude << ",\n"
            << "    \"longitude\" : " << closestStartNode->longitude << "\n"
            << "  },\n"
            << "  \"endNode\" : \n"
            << "  {\n"
            << "    \"latitude\" : "  << closestEndNode->latitude << ",\n"
            << "    \"longitude\" : " << closestEndNode->longitude << "\n"
            << "  },\n"
            << "  \"path\" : \n"
            << "  [";
      for (size_t i = 0; i < shortestPath->size(); i++)
      {
        if (i == 0)
        {
          jsonp << (*shortestPath)[i].latitude;
          jsonp << ", ";
          jsonp << (*shortestPath)[i].longitude;
          jsonp << ", ";
        }
        else if (i > 0) jsonp << ", ";
        jsonp << (*shortestPath)[i].latitude;
        jsonp << ", ";
        jsonp << (*shortestPath)[i].longitude;
      }
      jsonp << "]" << "\n";
      jsonp << "})\n";

      std::ostringstream answer;
      answer << "HTTP/1.0 200 OK\r\n"
             << "Content-Length: " << jsonp.str().size() << "\r\n"
             << "Content-Type: application/javascript" << "\r\n"
             << "Connection: close\r\n"
             << "\r\n"
             << jsonp.str();

      printf("\nAnswertext:  %s\n", answer.str().c_str());
      boost::system::error_code write_error;
      boost::asio::write(socket, boost::asio::buffer(answer.str()),
          boost::asio::transfer_all(), write_error);
    }
  }
  catch (std::exception& e) // NOLINT
  {
    std::cerr << e.what() << std::endl;
  }
}

